var LOGFILE_NAME = "fridge-supervision.log";
var l = new Logger(LOGFILE_NAME);
var verbose = 10;
var minResolution = 60;
var marginrate = 0.4;											//error tolerance
var consumptionMin = 32;											//minimum fridge consumption, learning output parameter
var consumptionMax = 109;											//maximum fridge consumption, learning output parameter, -1 indicates having a 2 device situation
var onTimeMin = 6;												//minimum time of consumption pulse, learning output parameter
var onTimeMax = 16;												//maximum time of consumption pulse, learning output parameter
var offTimeMin = 6;												//minimum idle time of fridge, learning output parameter
var offTimeMax = 16;												//maximum idle time of fridge, learning output parameter
var level=new Array(3);											//the identified 3 levels of consumption

numberSort = function (a,b) {									//to help built-in javascript sort
    return a - b;
};

function learning() {									//makes the algorithm adapt to fridge and compare later behaviors
	l.logln("Analyzing...");
	var Meters = getDSMeters();
	var oMeter = Meters[Property.getProperty("dSM_ID")];
	l.logln("Got meters.");		
    var rawdata = Metering.getValues(oMeter.dsid, "consumption", minResolution);		//get values
	l.logln("Got values. Example:"+rawdata[700].value);	
	var sorted_data = new Array();
	for (var i=0; i<rawdata.length; i++){
		sorted_data[i]=rawdata[i].value;						//copy data in new array
		rawdata[i]=sorted_data[i]
	}	
    sorted_data.sort(numberSort);								//sort them
	l.logln("Ascending consumption measured: "+JSON.stringify(sorted_data));	
	level[0]=sorted_data[~~(0.05*rawdata.length)];				//lowest consumption level
	level[1]=sorted_data[~~(0.75*rawdata.length)];				//high consumption level
	level[2]=sorted_data[~~(0.95*rawdata.length)];				//very high consumption level
	if (((level[2]-level[0])/(level[1]-level[0]))>1.5) {			//3 distinct levels of consumption. 2 fridges
		l.logln(" Two fridges detected?!")
		consumptionMax=-1;
		l.logln(" //level[0]: " + level[0] + " // level[1]: " + level[1] +" // level[2]: " + level[2]);
	}else{
		consumptionMin=(1-marginrate+0.0)*(sorted_data[~~(0.75*rawdata.length)]-sorted_data[~~(0.25*rawdata.length)]);	//get minimum consumption as the data[75%]-data[25%] with applied error margin
		consumptionMax=(1+marginrate+0.2)*(sorted_data[~~(0.75*rawdata.length)]-sorted_data[~~(0.25*rawdata.length)]);	//get maximum consumption as the data[75%]-data[25%] with applied error margin plus spike
		var Ton = new Array();										//variable initializations, all on-times
		var Toff = new Array();										//all off-times
		var timer = 0;												//a timer variable used as temp
		var ON_flag=false;											//true if we are on a +pulse, false if we are on a -pulse
		var ST_flag=false;											//indicates if the process has started. It will start with a positive pulse for reference reasons
		var Koff=0;													//temp indices of Ton and Toff
		var Kon=0;													//temp indices of Ton and Toff
		var W=0;													//temp energy variable
		for (var i=3; i<rawdata.length; i++){						//scan all data, starting from 3 because of preprocessing, checking the differential of each point(i) and previous point(i-3)
			W=rawdata[i]-rawdata[i-3];								//current consumption differencial
			if ((W<=consumptionMax)&&(W>=consumptionMin)){			//if it is positive and within limits
				if (ST_flag){										//and the process has started
					if (~ON_flag){									//if we are on Off-time
						Toff[Koff]=i-timer;							//mark the completion of this off-time and push it to array
						Koff++;										//pointing to next potential value
						ON_flag=true;								//now we are on On-time
					}
				} else {
					ST_flag=true;									//if we have not started yet, now set the start flag
					ON_flag=true;									//starting from positive pulse
				}
				timer=i;											//remember the time of detecting this edge
				//l.logln("UP at "+timer+ " with "+W+ "on rawdata" + rawdata[i]);
			}else if ((Math.abs(W)<=consumptionMax)&&(Math.abs(W)>=consumptionMin)&&(ST_flag)){ 	//if it is negative and within limits
				if (ON_flag) {										//if we are on On-time
					Ton[Kon]=i-timer;								//mark the completion of this on-time and push it to array
					Kon++;											//pointing to next potential value
					ON_flag=false;									//now we are on Off-time
				}
				timer=i;											//remember the time of detecting this edge
				//l.logln("DOWN at "+timer+ " with "+W+ "on rawdata" + rawdata[i]);				
			}	
		}
		l.logln("On times: "+JSON.stringify(Ton));
		l.logln("Off times: "+JSON.stringify(Toff));		
		Ton.sort(numberSort);										//ascending sort all on times
		Toff.sort(numberSort);										//ascending sort all off times		
		onTimeMax=(1+marginrate)*Ton[Kon-1];						//maximum on-time observed+safety margin
		offTimeMax=(1+marginrate)*Toff[Koff-1];						//maximum off-time observed+safety margin
		for (timer=1;timer<Kon;timer++){							//find the first time > 5minutes, to discard noise
			if (Ton[timer]>5) {break;}
		}
		onTimeMin=(1-marginrate)*Ton[~~((timer+Ton.length)/2)]		//the median of remaining on-times is the typical. Find minimum by applying safety margin
		for (timer=1;timer<Koff;timer++){							//find the first time > 5minutes, to discard noise
			if (Toff[timer]>5) {break;}
		}
		offTimeMin=(1-marginrate-0.2)*Toff[~~((timer+Toff.length)/2)]	//the median of remaining off-times is the typical. Find minimum by applying safety margin and 
														//considering that there are no negative spikes from other devices, so usually a negative pulse is our fridge
		l.logln(" // consumptionMin: " + consumptionMin + " // consumptionMax: " + consumptionMax +" // onTimeMin: " + onTimeMin +" // onTimeMax: " + onTimeMax +" // offTimeMin: " + offTimeMin +" // offTimeMax: " + offTimeMax);	
	}
	var MinC =Property.getProperty("MinC");
	var MaxC =Property.getProperty("MaxC");
	var MinTon =Property.getProperty("MinTon");
	var MaxTon =Property.getProperty("MaxTon");
	var MinToff =Property.getProperty("MinToff");
	var MaxToff =Property.getProperty("MaxToff");
	var Lev0 =Property.getProperty("Lev0");
	var Lev1 =Property.getProperty("Lev1");
	var Lev2 =Property.getProperty("Lev2");
	if ((MaxC==0)&&(Lev2==0)){
		l.logln("Parameter history is clear, setting parameters to current calculated...");
		Property.setProperty("MinC",Math.round(consumptionMin));	
		Property.setProperty("MaxC",Math.round(consumptionMax));
		Property.setProperty("MinTon",Math.round(onTimeMin));
		Property.setProperty("MaxTon",Math.round(onTimeMax));
		Property.setProperty("MinToff",Math.round(offTimeMin));
		Property.setProperty("MaxToff",Math.round(offTimeMax));
		Property.setProperty("Lev0",Math.round(level[0]));
		Property.setProperty("Lev1",Math.round(level[1]));	
		Property.setProperty("Lev2",Math.round(level[2]));	
	}else{
		l.logln("Parameter history is not clear, setting parameters to average of history and current calculated...");
		Property.setProperty("MinC",Math.round((MinC+consumptionMin)/2));	
		Property.setProperty("MaxC",Math.round((MaxC+consumptionMax)/2));
		Property.setProperty("MinTon",Math.round((MinTon+onTimeMin)/2));
		Property.setProperty("MaxTon",Math.round(Math.max(MaxTon,onTimeMax)));
		Property.setProperty("MinToff",Math.round((MinToff+offTimeMin)/2));
		Property.setProperty("MaxToff",Math.round(Math.max(MaxToff,offTimeMax)));
		Property.setProperty("Lev0",Math.round((Lev0+level[0])/2));
		Property.setProperty("Lev1",Math.round((Lev1+level[1])/2));	
		Property.setProperty("Lev2",Math.round((Lev2+level[2])/2));
	}
	var MinC =Property.getProperty("MinC");
	var MaxC =Property.getProperty("MaxC");
	var MinTon =Property.getProperty("MinTon");
	var MaxTon =Property.getProperty("MaxTon");
	var MinToff =Property.getProperty("MinToff");
	var MaxToff =Property.getProperty("MaxToff");
	var Lev0 =Property.getProperty("Lev0");
	var Lev1 =Property.getProperty("Lev1");
	var Lev2 =Property.getProperty("Lev2");	
	l.logln(" // consumptionMin: " + MinC + " // consumptionMax: " + MaxC +" // onTimeMin: " + MinTon +" // onTimeMax: " + MaxTon +" // offTimeMin: " + MinToff +" // offTimeMax: " + MaxToff);	
}

function preprocess(data) {										//used to switch from raw data to differential
	l.logln("Preprocessing...");	
	var processedArray = new Array();				//declare an array by 3 samples smaller
	for (var i=data.length-2; i > 1; i--) {						//run through each element starting from the most recent
		processedArray[data.length-2-i] = data[i-2].value-data[i+1].value;	//and keep its difference from the one 3 points ahead of it
	}	
	l.logln("	Done.");
	return processedArray;										//return the modified array
}

function getQualifiedEdge(data, i)	{							//returns the first valid edge starting scanning from position i
	l.logln(" Looking for edge...");
	var result = new Object();									//initialize a result
	result.edge = 0; 											//Zero means no edge found
	result.pos = data.length;									//initially set position on the end
	var edgeFound = false;										//flag for function stop
	while ((!edgeFound) && (i < data.length)) {					//while there are samples still remaining and no result was found
		if ((Math.abs(data[i]) >= consumptionMin) && (Math.abs(data[i]) <= consumptionMax)) {	// Qualifies, i.e. it's our fridge :-)
			if (data[i] > 0) {									// Rising edge -|_
				result.edge = +1;
			} else {											// Falling edge _|-
				result.edge = -1;
			}
			result.pos = i;									//its position
			edgeFound = true;									//stop searching
		}
		i++;													//next measurement
	}
	l.logln("	Found qualified edge at "+result.pos+" type: "+result.edge);	
	return result;												//return as object	
}

function getPulse(data, i)	{									//returns the first valid pulse if there are pulses on top of others																//or the first complete pulse regardless of duration validation, starting scanning from position i
	l.logln(" getPulse() has been called");	
	var result = new Object();									//initialize a result
	result.pulse=0;												//Zero means no pulse found
	result.ending=i;										//initially set position on the end
	result.duration=0;											//Zero means no pulse found	
	var firstEdge = getQualifiedEdge(data, i);					//identify the next valid edge
	result.ending = firstEdge.pos;								//it's the beginning of the pulse
	if (firstEdge.edge == 0) {									//failed to find fridge
		l.logln(" Failed to find valid edge, i= "+i);
		return result;											//return a zero pulse to indicate that
	}
	l.logln(" getPulse() found first edge...");		
	i = firstEdge.pos+3;									//keep searching after this point !!!SKIPPING 2 POINTS TO SIMPLIFY DEALING WITH SLOPES
	while (i < data.length) {		
		l.logln(" getPulse() looking for match...");	
		var secondEdge = getQualifiedEdge(data, i);				//identify the next valid edge (second)
		if (secondEdge.edge != firstEdge.edge) {				//if it's not the same kind as first, we have a pulse
			if (firstEdge.edge == 0) {									//failed to find fridge
				l.logln(" Failed to find valid edge, i= "+i);
				return result;											//return a zero pulse to indicate that
			}
			result.duration=secondEdge.pos-firstEdge.pos;	//its duration. The upper function calling us will decide if it's valid
			result.pulse=-1;									//set type of pulse as negative
			if (secondEdge.edge==-1) {result.pulse=+1;}				//change if it is not
			l.logln(" getPulse() found matching edge. Pulse ending in "+result.ending+". Returning.");			
			break;												//stop the while loop
		}
		i = secondEdge.pos;									//if it's the same kind, we have a pulse on top of pulse		
		l.logln(" getPulse() detected pulse at position " + i + " on top of other pulse, recursive scanning from position "+i+ " ...");				
		var temp_pulse=getPulse(data,i);						//use the last edge as first edge of the top pulse, and find the completion point
		if (((temp_pulse.duration>=onTimeMin)&&(temp_pulse.pulse==+1))||((temp_pulse.duration>=offTimeMin)&&(temp_pulse.pulse==-1))) { 
        return temp_pulse;
        l.logln(" valid pulse returning, duration: "+temp_pulse.duration+" ,type: "+temp_pulse.pulse+" ,ending at: "+temp_pulse.ending); 
        }		//qualified pulse. Return.
		i+=temp_pulse.duration;									//not qualified pulse. Keep searching for our second edge	
	}
	return result;												//return as object
}

function runAlgorithm(data) {									//the algorithm structure
// returns 	0 if fridge is OK
//			1 if fridge is dead
//			2 if fridge door is open
//			-1 if fridge could not be detected
	if (Property.getProperty("MaxC") > 0) {
		consumptionMin =Property.getProperty("MinC");
		consumptionMax =Property.getProperty("MaxC");
		onTimeMin =Property.getProperty("MinTon");
		onTimeMax =Property.getProperty("MaxTon");
		offTimeMin =Property.getProperty("MinToff");
		offTimeMax =Property.getProperty("MaxToff");
		level[0] =Property.getProperty("Lev0");
		level[1] =Property.getProperty("Lev1");
		level[2] =Property.getProperty("Lev2");	
	}
	//onTimeMax = 7;	
	//offTimeMax = 5;
	var i = 0;													//a variable that indicates position. Reference 0 is the most recent measurement of comsumption difference
	if (consumptionMax>-1){										//1 fridge
		while (i < data.length) {									//scan while still in limits
			l.logln(" Attempting to get a pulse from runAlgorithm()");
			var detectedPulse = getPulse(data,i);					//get a pulse
			detectedPulse.ending+=2;								//copensate for the differential preproccess
			l.logln(" Detected pulse, type: "+detectedPulse.pulse+", duration: "+ detectedPulse.duration + ", ending at "+detectedPulse.ending);
			if ((detectedPulse.duration>=onTimeMin)&&(detectedPulse.pulse==+1)){		//check if it is a valid positive pulse
				l.logln(" OFFTIME: "+detectedPulse.ending);
				l.logln("OFFTIME: "+detectedPulse.ending);
				Property.setProperty("OffTime",detectedPulse.ending);				
				if (detectedPulse.ending > offTimeMax) {				//Check if fridge is dead
					return 1;
				} else {
					return 0;					
				}
			}
			if ((detectedPulse.duration>=offTimeMin)&&(detectedPulse.pulse==-1)){		//check if it is a valid negative pulse
				l.logln(" ONTIME: "+detectedPulse.ending);
				l.logln("ONTIME: "+detectedPulse.ending);
				Property.setProperty("OnTime",detectedPulse.ending);			
				if (detectedPulse.ending > onTimeMax) {				//Check if door is open
					return 2;
				} else {
					return 0;
				}
			}
			i=detectedPulse.ending + detectedPulse.duration +1;
		}
		return -1;
	}else{															//2 fridges
		for (var k=data.length;k>(data.length-120);k--){			//2hour search for highest level
			if ((data[k-1]-level[0])>(level[2]-level[0])*(0.8)){break;}
			if (k==(data.length-119)){return 1;}
		}
		for (var k=data.length;k>(data.length-60);k--){			//1hour search for lowest level
			if (data[k-1]<level[0]*(1.2)){break;}
			if (k==(data.length-59)){return 2;}
		}	
		return 0;
	}
}

function resetlearning(){
	Property.setProperty("MinC",0);	
	Property.setProperty("MaxC",0);
	Property.setProperty("MinTon",0);
	Property.setProperty("MaxTon",0);
	Property.setProperty("MinToff",0);
	Property.setProperty("MaxToff",0);
	Property.setProperty("Lev0",0);
	Property.setProperty("Lev1",0);	
	Property.setProperty("Lev2",0);
	Property.setProperty("OnTime",0);
	Property.setProperty("OffTime",0);	
	var MinC =Property.getProperty("MinC");
	var MaxC =Property.getProperty("MaxC");
	var MinTon =Property.getProperty("MinTon");
	var MaxTon =Property.getProperty("MaxTon");
	var MinToff =Property.getProperty("MinToff");
	var MaxToff =Property.getProperty("MaxToff");
	var Lev0 =Property.getProperty("Lev0");
	var lev1 =Property.getProperty("Lev1");
	var Lev2 =Property.getProperty("Lev2");
	l.logln("Reseted learning: // consumptionMin: " + MinC + " // consumptionMax: " + MaxC +" // onTimeMin: " + MinTon +" // onTimeMax: " + MaxTon +" // offTimeMin: " + MinToff +" // offTimeMax: " + MaxToff);	
}
	
function uploadScheduler()
{
	l.logln("uploadScheduler started");
    var ecount = Property.getProperty('count');
	if (ecount!=2){
		var diff = 10 * 60;
		//l.logln("Previous timed event not found");		
		var uploadEvent = new TimedEvent("fridge-supervision", "+" + diff, { 'actions' : 'analyse' });
		var id = uploadEvent.raise();
		l.logln("Event set:" + id);		
		var date = new Date();
		//var evt = new Event("eventmailerexecute",{'to':MAILADDR,'body':'event scheduled','subject':date.toString(),'from':'dSS-Benachrichtigung'});			
		//evt.raise();																		//send mail
		l.logln("Scheduling done");
	}
	Property.setProperty("count",1);	
	l.logln("uploadScheduler completed");	
}

function analyze() {
	l.logln("Analyzing...");
	var Meters = getDSMeters();
	var oMeter = Meters[Property.getProperty("dSM_ID")];
	l.logln("Got meters.");		
    var meteringValues = Metering.getValues(oMeter.dsid, "consumption", minResolution);		//get values
	l.logln("Got values. Example:"+meteringValues[650].value);	
	var preprocessedData = preprocess(meteringValues);										//preprocess
	var result = runAlgorithm(preprocessedData);											//run algorithm
	l.logln("Algorithm result is: " + result);												//log
	var date = new Date();																	//get date
	Property.setProperty("lastRun", date.toString());										//log
	if (result > 0) {																		//if open door or malfunction
		var responseMSG;
		if (result==1) {responseMSG="Fridge Malfunction";}
		if (result==2) {responseMSG="Open Door";}
		var MAILADDR=Property.getProperty("MAILADDR");
		var evt = new Event("eventmailerexecute",{'to':MAILADDR,'body':responseMSG,'subject':date.toString(),'from':'dSS-Benachrichtigung'});			
		evt.raise();																		//send mail
		l.logln("mail sent");
	}
	uploadScheduler();	
}


function main() {
	l.logln("------------------got event" + raisedEvent.name+ " -------------------------");
    if (raisedEvent.name === "running") {
        Property.load();
    }else if (raisedEvent.name == 'model_ready') {
		var prop = new Property('count');
		prop.setValue(0);
	}else if (raisedEvent.name === "fridge-supervision") {
		l.logln("got action " + raisedEvent.parameter.actions);
        if (raisedEvent.parameter.actions) {
            if (verbose >= 1) {l.logln('Execute: ' + raisedEvent.name + ' ' + raisedEvent.parameter.actions);}
            if (raisedEvent.parameter.actions == 'load') { Property.load();}
            if (raisedEvent.parameter.actions == 'save') { Property.store();}
            if (raisedEvent.parameter.actions == 'reset') { resetlearning();}
            if (raisedEvent.parameter.actions == 'learn') { learning();}			
            if (raisedEvent.parameter.actions == 'prepare') { sendMeterData();}
            if (raisedEvent.parameter.actions == 'analyse') { analyze();}
        }
    }
}

main();

